
/*
 *  finddup is a simple utility find duplicated files, files common to
 *  several directories, or files present in one directory and not in
 *  another one.
 *
 *  Copyright (c) 2010 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of finddup.
 *
 *  finddup is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  finddup is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with finddup.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define VERSION_NUMBER "0.6"

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <getopt.h>
#include <fcntl.h>

#define BUFFER_SIZE 4096

typedef int64_t size_sum_t;

/* Yeah, global variables! */

int ignore_dotfiles = 0; /* 1 means ignore files and directories
                            starting with a dot */

int show_realpaths = 0; /* 1 means ignore files and directories
                            starting with a dot */

int show_progress = 1; /* 1 means show a progress bar when we are in a
                          tty */

int show_hits = 1; /* 1 means to show the files from dir2
                      corresponding to the ones from dir1 */

int show_groups = 1; /* 1 means to show the group IDs when printing
                        file names */

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if (!p && n != 0) {
    printf("Can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (ignore_dotfiles && name[0] == '.');
}

void print_size_sum(size_sum_t s) {
  char tmp[100];
  char *a = tmp + sizeof(tmp)/sizeof(char);
  *(--a) = '\0';
  if(s) {
    while(s) {
      *(--a) = s%10 + '0';
      s /= 10;
    }
  } else {
    *(--a) = '0';
  }
  printf(a);
}

/**********************************************************************/

struct file_with_size {
  char *filename;
  size_t size;
  ino_t inode;
  struct file_with_size *next;
  int id;
};

void file_list_delete(struct file_with_size *head) {
  struct file_with_size *next;
  while(head) {
    next = head->next;
    free(head->filename);
    free(head);
    head = next;
  }
}

int file_list_length(struct file_with_size *head) {
  int l = 0;
  while(head) {
    l++;
    head = head->next;
  }
  return l;
}

/**********************************************************************/

int same_content(struct file_with_size *f1, struct file_with_size *f2) {
  int fd1, fd2, s1, s2;
  char buffer1[BUFFER_SIZE], buffer2[BUFFER_SIZE];

  fd1 = open(f1->filename, O_RDONLY);
  fd2 = open(f2->filename, O_RDONLY);

  if(fd1 >= 0 && fd2 >= 0) {
    while(1) {
      s1 = read(fd1, buffer1, BUFFER_SIZE);
      s2 = read(fd2, buffer2, BUFFER_SIZE);

      if(s1 < 0 || s2 < 0) {
        close(fd1);
        close(fd2);
        return 0;
      }

      if(s1 == s2) {
        if(s1 == 0) {
          close(fd1);
          close(fd2);
          return 1;
        } else {
          if(strncmp(buffer1, buffer2, s1)) {
            close(fd1);
            close(fd2);
            return 0;
          }
        }
      } else {
        fprintf(stderr,
                "Different read size without error on files of same size.\n");
        exit(EXIT_FAILURE);
      }
    }
  } else {

    if(fd1 < 0) {
      fprintf(stderr,
              "Can not open \"%s\" error: %s\n",
              f1->filename,
              strerror(errno));
    }

    if(fd2 < 0) {
      fprintf(stderr,
              "Can not open \"%s\" error: %s\n",
              f2->filename,
              strerror(errno));
    }
    exit(EXIT_FAILURE);
  }
}

int same_files(struct file_with_size *f1, struct file_with_size *f2) {
  return f1->size == f2->size && same_content(f1, f2);
}

/**********************************************************************/

struct file_with_size *scan_directory(struct file_with_size *tail,
                                      const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat sb;
  struct file_with_size *tmp;
  char subname[PATH_MAX];

  if(lstat(name, &sb) != 0) {
    fprintf(stderr, "Can not stat \"%s\": %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(sb.st_mode)) {
    return tail;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", name, dir_e->d_name);
        tail = scan_directory(tail, subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(sb.st_mode)) {
      tmp = safe_malloc(sizeof(struct file_with_size));
      tmp->next = tail;
      tmp->filename = strdup(name);
      tmp->size = sb.st_size;
      tmp->inode = sb.st_ino;
      tmp->id = -1;
      tail = tmp;
    }
  }

  return tail;
}

void print_file(struct file_with_size *node) {
  if(show_realpaths) {
    if(show_groups) {
      printf("%d %s\n", node->id, realpath(node->filename, 0));
    } else {
      printf("%s\n", realpath(node->filename, 0));
    }
  } else {
    if(show_groups) {
      printf("%d %s\n", node->id, node->filename);
    } else {
      printf("%s\n", node->filename);
    }
  }
}

void start(const char *dirname1, const char *dirname2) {
  struct file_with_size *list1, *list2;
  struct file_with_size *node1, *node2;
  struct stat sb1, sb2;
  int not_in, found, same_dir;

  if(strncmp(dirname2, "not:", 4) == 0) {
    not_in = 1;
    dirname2 += 4;
  } else {
    not_in = 0;
  }

  if(lstat(dirname1, &sb1) != 0) {
    fprintf(stderr, "Can not stat \"%s\": %s\n", dirname1, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(lstat(dirname2, &sb2) != 0) {
    fprintf(stderr, "Can not stat \"%s\": %s\n", dirname2, strerror(errno));
    exit(EXIT_FAILURE);
  }

  same_dir = (sb1.st_ino == sb2.st_ino);

  fprintf(stderr, "Scanning %s ... ", dirname1);
  list1 = scan_directory(0, dirname1);
  if(same_dir) {
    list2 = list1;
  } else {
    fprintf(stderr, "%s ... ", dirname2);
    list2 = scan_directory(0, dirname2);
  }
  fprintf(stderr, "done.\n");

  if(not_in) {
    for(node1 = list1; node1; node1 = node1->next) {
      found = 0;

      for(node2 = list2; !found && node2; node2 = node2->next) {
        if(node1->inode != node2->inode && same_files(node1, node2)) {
          found = 1;
        }
      }

      if(!found) {
        if(show_realpaths) {
          printf("%s\n", realpath(node1->filename, 0));
        } else {
          printf("%s\n", node1->filename);
        }
      }
    }

  } else {
    int k = 0;

    for(node1 = list1; node1; node1 = node1->next) {
      for(node2 = list2; node2; node2 = node2->next) {
        if(node1->inode != node2->inode && same_files(node1, node2)) {
          if(node1->id < 0) {
            if(node2->id >= 0) {
              node1->id = node2->id;
            } else {
              node1->id = k;
              k++;
            }
            print_file(node1);
          }
          if(node2->id < 0) {
            node2->id = node1->id;
            if(show_hits) {
              print_file(node2);
            }
          }
        }
      }
    }
  }

  file_list_delete(list1);
  if(!same_dir) {
    file_list_delete(list2);
  }
}

void print_help(FILE *out) {
  fprintf(out, "Usage: finddup [OPTION]... DIR1 [[not:]DIR2]\n");
  fprintf(out, "Version %s (%s)\n", VERSION_NUMBER, UNAME);
  fprintf(out, "Without DIR2, lists duplicated files found in DIR1. With DIR2, lists files common to both directories. With the not: prefix, lists files found in DIR1 which do not exist in DIR2.\n");
  fprintf(out, "\n");
  fprintf(out, "   -h   show this help\n");
  fprintf(out, "   -d   ignore dot files and directories\n");
  fprintf(out, "   -c   do not show which files in DIR2 corresponds to those in DIR1\n");
  fprintf(out, "   -g   do not show the file groups\n");
  fprintf(out, "   -r   show the real file paths\n");
  fprintf(out, "\n");
  fprintf(out, "Report bugs and comments to <francois@fleuret.org>\n");
}

/**********************************************************************/

int main(int argc, char **argv) {
  int c;
  struct file_with_size *root;

  root = 0;

  setlocale (LC_ALL, "");

  while (1) {
    c = getopt(argc, argv, "hrcgd");
    if (c == -1)
      break;

    switch (c) {

    case 'h':
      print_help(stdout);
      exit(EXIT_SUCCESS);

      break;

    case 'd':
      ignore_dotfiles = 1;
      break;

    case 'r':
      show_realpaths = 1;
      break;

    case 'g':
      show_groups = 0;
      break;

    case 'c':
      show_hits = 0;
      break;

    default:
      exit(EXIT_FAILURE);
    }
  }

  if(optind + 1 < argc) {
    start(argv[optind], argv[optind + 1]);
  } else if(optind < argc) {
    start(argv[optind], argv[optind]);
  } else {
    print_help(stderr);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);
}
